#!/usr/bin/env python3
# -*- coding:utf-8 -*-

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# Example:
# ./execute.py United_Kingdom 3 2.0 -1.2 5.0 1 1.0 21 Ma

"""
(c) 2020 Miguel Colom
http://mcolom.info

Under GNU GPL license. See LICENCE file for details.
"""

import sys
import os
import urllib.request
import datetime
import time
import argparse
import json
from numpy import roll
import pandas as pd

def demo_failure(message):
    '''
    Exit with a demo_failure message
    '''
    print(message)
    with open("demo_failure.txt", "wt") as f:
        f.write(message)
        f.write("\n")
    sys.exit(0)

def get_today_str():
    '''
    Return a string representing today's date
    '''
    today = datetime.date.today()
    year, month, day = today.year, today.month, today.day
    return f"{year}-{month}-{day}"

def get_cached_json_filename(data_directory):
    '''
    Get the path of the cached JSON file with all countries's data
    '''
    today_str = get_today_str()
    return os.path.join(data_directory, f"data_{today_str}.json")

def get_cached_filename_USA(data_directory):
    '''
    Get the path of the cached USA's file with all state's data
    '''
    today_str = get_today_str()
    return os.path.join(data_directory, f"USA_{today_str}.csv")

def get_records_by_state_USA(df):
    '''
    Structure the cases and dates of all countries
    '''
    report_dates = {}
    report_cases = {}
    report_deaths = {}

    states = df['state'].unique().astype(str)
    for state_spaces in states:
        state = state_spaces.replace(' ', '_').replace(',', '_')

        # Cases
        data = [0 if pd.isna(cases) else int(cases) for cases in df[df['state'] == state_spaces]['cases']]
        data = (data - roll(data, 1)) # De-accumulate
        report_cases[state] = data[1:]

        # Deaths
        data = [0 if pd.isna(deaths) else int(deaths) for deaths in df[df['state'] == state_spaces]['deaths']]
        data = (data - roll(data, 1)) # De-accumulate
        report_deaths[state] = data[1:]

        # Dates
        data = [pd.to_datetime(timestamp) for timestamp in df[df['state'] == state_spaces]['date']]
        report_dates[state] = data[1:]


    return report_dates, report_cases, report_deaths

def get_records_by_country_json(records):
    '''
    Structure the cases and dates of all countries
    '''
    report_dates = {}
    report_populations = {}
    report_cases = {}
    report_deaths = {}

    for record in records:
        country = record['countriesAndTerritories']
        populations = int(record['popData2019']) if record['popData2019'] else None
        cases = int(record['cases'])
        deaths = int(record['deaths'])

        date_str = record['dateRep']
        dt = datetime.datetime.strptime(date_str, '%d/%m/%Y')

        if not country in report_dates:
            report_dates[country] = []
        report_dates[country].append(dt)

        if not country in report_populations:
            report_populations[country] = []
        report_populations[country].append(populations)

        if not country in report_deaths:
            report_deaths[country] = []
        report_deaths[country].append(deaths)

        if not country in report_cases:
            report_cases[country] = []
        report_cases[country].append(cases)

    return report_dates, report_populations, report_cases, report_deaths

def save_country(cases, deaths, start_date, population, country, data_directory):
    '''
    Save the cases of one particular country
    '''
    filename = os.path.join(data_directory, f"cases_{country}.txt")
    with open(filename, "wt") as f:
        for num_cases in cases:
            f.write(f"{num_cases}\n")

    filename = os.path.join(data_directory, f"deaths_{country}.txt")
    with open(filename, "wt") as f:
        for num_deaths in deaths:
            f.write(f"{num_deaths}\n")

    filename = os.path.join(data_directory, f"start_date_{country}.txt")
    with open(filename, "wt") as f:
        f.write(f"{start_date.timestamp()}\n")

    filename = os.path.join(data_directory, f"population_{country}.txt")
    with open(filename, "wt") as f:
        f.write(f"{population}\n")

def acquire_lock(lock_path):
    """
    Create data write lock in order to prevent concurrent writes
    """
    open(lock_path, 'w+').close()

def is_locked(lock_path):
    """
    Check if there's a locked data write
    """
    if os.path.isfile(lock_path):
        current_time = time.time()
        creation_time = os.path.getctime(lock_path)
        # In case of error it might leave dangling lockfiles, remove it if it's old.
        if current_time - creation_time >= 20*60:
            os.remove(lock_path)
            return False
        return True
    return False

def ensure_countries_data_updated_ECDC(data_directory):
    '''
    Ensure the ECDC's data of all countries is updated up to today
    '''
    # Check if today's data has been already downloaded less than the
    # specified number of hours ago. Return cached copy if so.
    filename = get_cached_json_filename(data_directory)

    # Read cached data
    if os.path.isfile(filename) and get_age_hours(filename) < 3:
        return

    # Acquire a lock, to prevent concurrent writes
    lock_path = os.path.join(data_directory, ".lock")
    if is_locked(lock_path):
        return
    acquire_lock(lock_path)

    # Download ECDC data with all countries
    # From ECDC directly:
    #url = "https://opendata.ecdc.europa.eu/covid19/casedistribution/json"
    # CTIM proxy:
    url = "http://ctimserver.dis.ulpgc.es/covid19/casedistribution/json"
    print(f"Downloading from {url}")

    try:
        response = urllib.request.urlopen(url, timeout=15).read()
    except urllib.error.URLError:
        demo_failure("We couldn't obtain the data since ECDC isn't responding")

    # Save downloaded file
    response_decoded = response.decode()
    with open(filename, "wt") as f:
        f.write(response_decoded)

    records = json.loads(response_decoded)['records']
    report_dates, report_populations, report_cases, report_deaths = get_records_by_country_json(records)

    for country in report_dates:
        start_date, end_date = min(report_dates[country]), max(report_dates[country])
        all_populations = []
        all_cases = []
        all_deaths = []

        # Go through all cases, from the start to the end dates
        date = start_date
        while date <= end_date:
            try:
                date_idx = report_dates[country].index(date)
            except ValueError:
                date_idx = None

            # Assume 0 cases if not given in the report
            num_populations = 0 if date_idx is None else report_populations[country][date_idx]
            num_cases = 0 if date_idx is None else report_cases[country][date_idx]
            num_deaths = 0 if date_idx is None else report_deaths[country][date_idx]
            all_populations.append(num_populations)
            all_cases.append(num_cases)
            all_deaths.append(num_deaths)

            # Next day
            date += datetime.timedelta(days=1)
        save_country(all_cases, all_deaths, start_date, all_populations[0], country, data_directory)

    if os.path.isfile(lock_path):
        os.remove(lock_path)

def ensure_countries_data_updated_USA(data_directory):
    '''
    Ensure the USA's data of all countries is updated up to today
    '''
    # Check if today's data has been already downloaded less than the
    # specified number of hours ago. If so, nothing to be done
    filename = get_cached_filename_USA(data_directory)
    #
    if os.path.isfile(filename) and get_age_hours(filename) < 3:
        return


    # Acquire a lock, to prevent concurrent writes
    lock_path = os.path.join(data_directory, ".lock_USA")
    if is_locked(lock_path):
        return
    acquire_lock(lock_path)

    # Download USA data for all states
    url = "https://raw.githubusercontent.com/nytimes/covid-19-data/master/us-states.csv"
    print(f"Downloading from {url}")

    try:
        response = urllib.request.urlopen(url, timeout=15).read()
    except urllib.error.URLError:
        demo_failure("We couldn't obtain the USA's data since GitHub isn't responding")

    # Save downloaded file
    with open(filename, "wb") as f:
        f.write(response)

    df = pd.read_csv(filename)
    report_dates, report_cases, report_deaths = get_records_by_state_USA(df)

    for state in report_dates:
        start_date, end_date = min(report_dates[state]), max(report_dates[state])
        all_cases = []
        all_deaths = []

        # Go through all cases, from the start to the end dates
        date = start_date
        while date <= end_date:
            try:
                date_idx = report_dates[state].index(date)
            except ValueError:
                date_idx = None

            # Assume 0 cases if not given in the report
            num_cases = 0 if date_idx is None else report_cases[state][date_idx]
            num_deaths = 0 if date_idx is None else report_deaths[state][date_idx]
            all_cases.append(num_cases)
            all_deaths.append(num_deaths)

            # Next day
            date += datetime.timedelta(days=1)
        save_country(all_cases, all_deaths, start_date, -1, state, data_directory)

    if os.path.isfile(lock_path):
        os.remove(lock_path)

def ensure_countries_data_updated(data_directory):
    '''
    Ensure the data of all countries is updated up to today
    '''
    # Read ECDC data
    ensure_countries_data_updated_ECDC(data_directory)

    # Read USA states data
    ensure_countries_data_updated_USA(data_directory)

def get_age_hours(filename):
    '''
    Get the age (modification time) of the file in hours
    '''
    file_dt = datetime.datetime.fromtimestamp(os.path.getmtime(filename))
    diff = datetime.datetime.now() - file_dt
    return diff.seconds / 3600


# Parse arguments
parser = argparse.ArgumentParser()
parser.add_argument("--data_directory", help="Data directory", default="gi_extras/storage")
args = parser.parse_args()

# Ensure all country data is updated
ensure_countries_data_updated(args.data_directory)
